﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Management;
using System.Windows.Forms;
using System.ComponentModel;
using System.Threading;
using System.Runtime.InteropServices;
using System.Diagnostics;

namespace RoboclawClassLib
{
    class USBDeviceInfo
    {
        public USBDeviceInfo(string name, string deviceID, string pnpDeviceID, string description)
        {
            this.Name = name;
            this.DeviceID = deviceID;
            this.PnpDeviceID = pnpDeviceID;
            this.Description = description;
        }
        public string Name { get; private set; }
        public string DeviceID { get; private set; }
        public string PnpDeviceID { get; private set; }
        public string Description { get; private set; }
    }

    public class Roboclaw
    {
        private System.IO.Ports.SerialPort serialPort;

		private bool m_ack;
		private byte m_address;

        const uint DIGCF_PRESENT = 0x02;
        const uint DIGCF_ALLCLASSES = 0x04;
        const uint ERROR_INVALID_DATA = 13;
        const uint ERROR_NO_MORE_ITEMS = 259;

        static DEVPROPKEY DEVPKEY_Device_BusReportedDeviceDesc;
        static DEVPROPKEY DEVPKEY_Device_HardwareIds;
        static DEVPROPKEY DEVPKEY_Device_FriendlyName;

        [StructLayout(LayoutKind.Sequential)]
        struct SP_CLASSINSTALL_HEADER
        {
            public UInt32 cbSize;
            public UInt32 InstallFunction;
        }

        [StructLayout(LayoutKind.Sequential)]
        struct SP_PROPCHANGE_PARAMS
        {
            public SP_CLASSINSTALL_HEADER ClassInstallHeader;
            public UInt32 StateChange;
            public UInt32 Scope;
            public UInt32 HwProfile;
        }

        [StructLayout(LayoutKind.Sequential)]
        struct SP_DEVINFO_DATA
        {
            public UInt32 cbSize;
            public Guid classGuid;
            public UInt32 devInst;
            public IntPtr reserved;     // CHANGE #1 - was UInt32
        }

        [StructLayout(LayoutKind.Sequential)]
        struct DEVPROPKEY
        {
            public Guid fmtid;
            public UInt32 pid;
        }

        [DllImport("setupapi.dll", SetLastError = true)]
        static extern IntPtr SetupDiGetClassDevsW(
            [In] ref Guid ClassGuid,
            [MarshalAs(UnmanagedType.LPWStr)] string Enumerator,
            IntPtr parent,
            UInt32 flags);

        [DllImport("setupapi.dll", SetLastError = true)]
        static extern bool SetupDiDestroyDeviceInfoList(IntPtr handle);

        [DllImport("setupapi.dll", SetLastError = true)]
        static extern bool SetupDiEnumDeviceInfo(IntPtr deviceInfoSet,
            UInt32 memberIndex,
            [Out] out SP_DEVINFO_DATA deviceInfoData);

        [DllImport("setupapi.dll", SetLastError = true)]
        static extern bool SetupDiSetClassInstallParams(
            IntPtr deviceInfoSet,
            [In] ref SP_DEVINFO_DATA deviceInfoData,
            [In] ref SP_PROPCHANGE_PARAMS classInstallParams,
            UInt32 ClassInstallParamsSize);

        [DllImport("setupapi.dll", SetLastError = true)]
        static extern bool SetupDiChangeState(
            IntPtr deviceInfoSet,
            [In] ref SP_DEVINFO_DATA deviceInfoData);

        [DllImport("setupapi.dll", SetLastError = true)]
        static extern bool SetupDiGetDevicePropertyW(
                IntPtr deviceInfoSet,
                [In] ref SP_DEVINFO_DATA DeviceInfoData,
                [In] ref DEVPROPKEY propertyKey,
                [Out] out UInt32 propertyType,
                IntPtr propertyBuffer,
                UInt32 propertyBufferSize,
                out UInt32 requiredSize,
                UInt32 flags);

        [DllImport("setupapi.dll", SetLastError = true)]
        static extern bool SetupDiGetDeviceRegistryPropertyW(
          IntPtr DeviceInfoSet,
          [In] ref SP_DEVINFO_DATA DeviceInfoData,
          UInt32 Property,
          [Out] out UInt32 PropertyRegDataType,
          IntPtr PropertyBuffer,
          UInt32 PropertyBufferSize,
          [In, Out] ref UInt32 RequiredSize
        );

        private static string GetStringPropertyForDevice(IntPtr info, SP_DEVINFO_DATA devdata,
            DEVPROPKEY propKey)
        {
            uint proptype, outsize;
            IntPtr buffer = IntPtr.Zero;
            try
            {
                uint buflen = 512;
                buffer = Marshal.AllocHGlobal((int)buflen);
                outsize = 0;
                SetupDiGetDevicePropertyW(
                info,
                ref devdata,
                ref propKey,
                out proptype,
                buffer,
                buflen,
                out outsize,
                0);
                
                byte[] lbuffer = new byte[outsize];
                Marshal.Copy(buffer, lbuffer, 0, (int)outsize);
                int errcode = Marshal.GetLastWin32Error();
                if (errcode == ERROR_INVALID_DATA) return null;
                return Encoding.Unicode.GetString(lbuffer);
            }
            finally
            {
                if (buffer != IntPtr.Zero)
                    Marshal.FreeHGlobal(buffer);
            }
        }
        
        static List<USBDeviceInfo> GetUSBDevices()
        {

            List<USBDeviceInfo> devices = new List<USBDeviceInfo>();

            DEVPKEY_Device_BusReportedDeviceDesc = new DEVPROPKEY();
            DEVPKEY_Device_BusReportedDeviceDesc.fmtid = new Guid(
                0x540b947e, 0x8b40, 0x45bc, 0xa8, 0xa2, 0x6a,
                0x0b, 0x89, 0x4c, 0xbd, 0xa2);
            DEVPKEY_Device_BusReportedDeviceDesc.pid = 4;

            DEVPKEY_Device_HardwareIds = new DEVPROPKEY();
            DEVPKEY_Device_HardwareIds.fmtid = new Guid(
                0xa45c254e, 0xdf1c, 0x4efd, 0x80, 0x20, 0x67,
                0xd1, 0x46, 0xa8, 0x50, 0xe0);
            DEVPKEY_Device_HardwareIds.pid = 3;

            DEVPKEY_Device_FriendlyName = new DEVPROPKEY();
            DEVPKEY_Device_FriendlyName.fmtid = new Guid(
                0xa45c254e, 0xdf1c, 0x4efd, 0x80, 0x20, 0x67,
                0xd1, 0x46, 0xa8, 0x50, 0xe0);
            DEVPKEY_Device_FriendlyName.pid = 14;

            IntPtr info = IntPtr.Zero;
            Guid guid = Guid.Empty;
            try
            {
                info = SetupDiGetClassDevsW(
                    ref guid,
                    null,
                    IntPtr.Zero,
                    DIGCF_ALLCLASSES | DIGCF_PRESENT);

                SP_DEVINFO_DATA devdata = new SP_DEVINFO_DATA();
                devdata.cbSize = (UInt32)Marshal.SizeOf(devdata);

                // Get first device matching device criterion.
                for (uint i = 0; ; i++)
                {
                    SetupDiEnumDeviceInfo(info,
                        i,
                        out devdata);
                    // if no items match filter, throw
                    if (Marshal.GetLastWin32Error() != ERROR_NO_MORE_ITEMS)
                    {
                        string busdescription = GetStringPropertyForDevice(info,
                                                 devdata, DEVPKEY_Device_BusReportedDeviceDesc);
                        if (busdescription == "RC15A32KULDR\0" ||
                            busdescription == "RC30A32KULDR\0" ||
                            busdescription == "RC60A32KULDR\0" ||
							busdescription == "RC15A64KULDR\0" ||
							busdescription == "RC30A64KULDR\0" ||
							busdescription == "RC60A64KULDR\0" ||
							busdescription == "RC60HV6032KULDR\0" ||
							busdescription == "RC60HV6064KULDR\0" ||
							busdescription == "RC60HV8032KULDR\0" ||
							busdescription == "RC60HV8064KULDR\0" ||
							busdescription == "RC80HV6032KULDR\0" ||
							busdescription == "RC80HV6064KULDR\0" ||
							busdescription == "RC150HV60ULDR\0" ||
							busdescription == "RC150HV80ULDR\0" ||
							busdescription == "GA40064KULDR\0" ||
							busdescription == "USB Roboclaw 2x15A\0" ||
                            busdescription == "USB Roboclaw 2x30A\0" ||
                            busdescription == "USB Roboclaw 2x60A\0" ||
							busdescription == "USB Roboclaw 2x60HV60\0" ||
							busdescription == "USB Roboclaw 2x60HV80\0" ||
							busdescription == "USB Roboclaw 2x80HV60\0" ||
							busdescription == "USB Roboclaw 2x150HV60\0" ||
							busdescription == "USB Roboclaw 2x150HV80\0" ||
							busdescription == "GA400 Stepper Control\0")
                        {
                            string hardwareids = GetStringPropertyForDevice(info,
                                                     devdata, DEVPKEY_Device_HardwareIds);
                            //Find just Atmel CDC Virtual Comports
                            if (hardwareids.Substring(0, 21) == "USB\\VID_03EB&PID_2404")
                            {
                                string friendly = GetStringPropertyForDevice(info,devdata, DEVPKEY_Device_FriendlyName);
								System.Text.RegularExpressions.Regex rgx = new System.Text.RegularExpressions.Regex(@"\((COM\d+)\)");  // "..... (COMxxx)" -> COMxxxx
								if (rgx.Matches(friendly).Count > 0){
									string raw = rgx.Match(friendly).ToString();
									System.Text.RegularExpressions.Regex rgx2 = new System.Text.RegularExpressions.Regex(@"(COM\d+)");  // "..... (COMxxx)" -> COMxxxx
									string comport = rgx2.Match(raw).ToString();
	                                devices.Add(new USBDeviceInfo(
		                                friendly,
			                            comport,
				                        hardwareids,
					                    busdescription));
								}
                            }
                        }
                    }
                    else
                        break;
                }
			}
            finally
            {
                if (info != IntPtr.Zero)
                    SetupDiDestroyDeviceInfoList(info);
            }


            return devices;
        }

        private bool Write_CRC(byte[] data)
        {
            string str = "";
            byte crc = 0;
            int len = data.GetLength(0);
            for (int i = 0; i < len; i++)
            {
                crc += data[i];
                str += string.Format("{0:X2} ", data[i]);
            }
            data[data.GetLength(0) - 1] = (byte)(crc & 0x7F);
            if (m_ack)
                data[data.GetLength(0) - 1] |= 0x80;
            str += string.Format("{0:X2} ", data[data.GetLength(0) - 1]);
            try
            {
                serialPort.Write(data, 0, len);
            }
            catch (Exception pEx)
            {
				Debug.WriteLine("Write_CRC Exception:" + pEx.Message);
				return false;
            }
            if (m_ack)
            {
                try
                {
                    serialPort.ReadByte();
                }
                catch (TimeoutException)
                {
					Debug.WriteLine("Write_CRC ACK Timeout"); 
					System.Threading.Thread.Sleep(20);
                    return false;
                }
            }
			return true;
        }

        private bool readcmd(byte address, byte cmd, uint len, out byte[] arr)
        {
			byte crc,ccrc;
			string str = "";
			arr = new byte[len];
			uint timeout = 0;
						
			do{
				if(timeout!=0)
					System.Threading.Thread.Sleep(100);

				timeout++;

				uint timeout0 = 0;
				do{
					serialPort.DiscardInBuffer();
					if (timeout0 != 0)
						System.Threading.Thread.Sleep(20);

					crc = address;
					crc += cmd;

					try{
						serialPort.Write(new byte[] { address }, 0, 1);
						str += string.Format("{0:X2} ", address);
					}
					catch (Exception pEx){
						Debug.WriteLine("ReadCmd Write Exception:" + pEx.Message);
						return false;
					}

					try{
						serialPort.Write(new byte[] { cmd }, 0, 1);
						str += string.Format("{0:X2} ", cmd);
					}
					catch (Exception pEx){
						Debug.WriteLine("ReadCmd Write Exception:" + pEx.Message);
						return false;
					}

					for (Int32 i = 0; i < len; i++)
					{
						try{
							arr[i] = (byte)serialPort.ReadByte();
							timeout0 = 0;
							crc += arr[i];
							str += string.Format("{0:X2} ", arr[i]);
						}
						catch (TimeoutException){
							serialPort.DiscardInBuffer();
							if (i == 0){
								timeout0++;
								break;
							}
							else{
								Debug.WriteLine("ReadCmd Data Timeout:" + cmd.ToString("X2") + " " + i.ToString() + " of " + len.ToString());
								return false;
							}
						}
					}
				}while(timeout0>0 && timeout0<5);
				if(timeout0==5){
					Debug.WriteLine("ReadCmd Timeout0 Retry Failure");
					return false;
				}

				try{
					ccrc = (byte)serialPort.ReadByte();
					str += string.Format("{0:X2} ", ccrc);
				}
				catch (TimeoutException){
					Debug.WriteLine("ReadCmd CRC Timeout:" + cmd.ToString("X2"));
					serialPort.DiscardInBuffer();
					return false;
				}
			}while(timeout<5 && (crc & 0x7F)!=ccrc);

			if ((crc & 0x7F) != ccrc){
				Debug.WriteLine("ReadCmd Invalid CRC:" + cmd.ToString("X2") + " " + str + " " + crc.ToString("X2") + " " + ccrc.ToString("X2"));
				serialPort.DiscardInBuffer();
				return false;
            }

            return true;
        }

        public bool IsOpen()
        {
            try
            {
				if(null!=serialPort)
					return serialPort.IsOpen;
				return false;
            }
            catch (Exception)
            {
				//MessageBox.Show(pEx.Message, "Error");
				return false;
            }
        }

        public void Close()
        {
            try
            {
                if (IsOpen())
                    serialPort.Close();
				serialPort.Dispose();
				serialPort=null;
            }
            catch (Exception)
            {
				//MessageBox.Show(pEx.Message, "Error");
			}
        }

        public string Open(
            string comport,
            ref string model,
            bool ack,
			byte address,
            Int32 baudrate)
        {
            m_ack = ack;
			m_address = address;

            try
            {
                if (comport == "AUTO")
                {
                    //Close port if its already open
                    if (serialPort != null)
                        if (serialPort.IsOpen)
                            serialPort.Close();

                    //This code will autodetect USB Roboclaw boards
                    var usbDevices = GetUSBDevices();   //Get all USB to Serial adapters
                    foreach (var usbDevice in usbDevices)
                    {
                        //Check if this is a USB Roboclaw
                        comport = usbDevice.DeviceID;
                        model = usbDevice.Description;
                    }

                    baudrate = 1000000;

                    if (comport == "AUTO")
                        return "";
                }

                if (serialPort == null)
                {
                    serialPort = new System.IO.Ports.SerialPort(comport, baudrate);
                    if (serialPort == null)
                        return "";
                }
                else
                {
                    if (serialPort.IsOpen)
                        serialPort.Close();
                    serialPort.PortName = comport;
                    serialPort.BaudRate = baudrate;
                }

                serialPort.Open();
                serialPort.ReadTimeout = 100;

                return comport;
            }
            catch (Exception)
            {
                //MessageBox.Show(pEx.Message,"Error");
                return "";
            }
        }

/*        private void TXbyte(byte one)
        {
            byte[] data = new byte[1];
            data[0] = one;
            serialPort.Write(
                data,
                0,
                1);
        }*/

		BackgroundWorker fw = new BackgroundWorker();
		
		private void fw_DoWork(object sender, DoWorkEventArgs e)
		{
			string model = "";
			if (Open("AUTO", ref model, true, 0x80, 0) == "")
			{
//				MessageBox.Show("Unable to open communications port", "Update Error");
				return;
			}

			int size = 0;
			string path = System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location);

			string fileName = "";
			if (model == "RC15A32KULDR\0"){
				fileName = path + "\\MC15A32K.bin";
				size = 0x8000;
			}
			else if (model == "RC15A64KULDR\0"){
				fileName = path + "\\MC15A64K.bin";
				size = 0x10000;
			}
			else if (model == "RC30A32KULDR\0"){
				fileName = path + "\\MC30A32K.bin";
				size = 0x8000;
			}
			else if (model == "RC30A64KULDR\0"){
				fileName = path + "\\MC30A64K.bin";
				size = 0x10000;
			}
			else if (model == "RC60A32KULDR\0"){
				fileName = path + "\\MC60A32K.bin";
				size = 0x8000;
			}
			else if (model == "RC60A64KULDR\0"){
				fileName = path + "\\MC60A64K.bin";
				size = 0x10000;
			}
			else if (model == "RC60HV6032KULDR\0"){
				fileName = path + "\\MC60HV6032K.bin";
				size = 0x8000;
			}
			else if (model == "RC60HV6064KULDR\0")
			{
				fileName = path + "\\MC60HV6064K.bin";
				size = 0x10000;
			}
			else if (model == "RC60HV8032KULDR\0")
			{
				fileName = path + "\\MC60HV8032K.bin";
				size = 0x8000;
			}
			else if (model == "RC60HV8064KULDR\0")
			{
				fileName = path + "\\MC60HV8064K.bin";
				size = 0x10000;
			}
			else if (model == "RC80HV6032KULDR\0"){
				fileName = path + "\\MC80HV6032K.bin";
				size = 0x8000;
			}
			else if (model == "RC80HV6064KULDR\0")
			{
				fileName = path + "\\MC80HV6064K.bin";
				size = 0x10000;
			}
			else if (model == "RC150HV60ULDR\0")
			{
				fileName = path + "\\MC150HV60.bin";
				size = 0x10000;
			}
			else if (model == "RC150HV80ULDR\0"){
				fileName = path + "\\MC150HV80.bin";
				size = 0x10000;
			}
			else{
				Close();
//				MessageBox.Show("Mode is incorrect or model is not supported", "Update Error");
				return;
			}

			if (File.Exists(fileName))
			{

				byte[] data = new byte[0x10000];
				using (BinaryReader reader = new BinaryReader(File.Open(fileName, FileMode.Open)))
				{
					reader.Read(data, 0, size);

					try
					{
						serialPort.ReadTimeout = 1000;
						//TXbyte(0x01);
						serialPort.Write(new byte[] { 0x01 }, 0, 1);
						int index = 0;
						for (uint c1 = 0; c1 < size; c1 += 256)
						{
							if (serialPort.ReadByte() != 0xaa)
							{
								MessageBox.Show("Error Writing: " + (c1 * 2).ToString(), "Update Error");
								Close();
								return;
							}
							for (int c2 = 0; c2 < 256; c2++)
							{
								//TXbyte(data[index++]);
								serialPort.Write(data, index++, 1);

								if ((c2 % 16) == 0)
								{
									fw.ReportProgress((int)(100 * (c1 + c2) / size));
								}
							}
						}

						if (serialPort.ReadByte() != 0xaa)
						{
							MessageBox.Show("Error Writing", "Update Error");
							Close();
							return;
						}

						//TXbyte(0x04);	//reset
						serialPort.Write(new byte[] { 0x04 }, 0, 1);
						serialPort.ReadTimeout = 100;
					}
					catch (Exception pEx)
					{
						MessageBox.Show("Unexpected communications error" + pEx.ToString(), "Update Error");
						Close();
						return;
					}
				}
			}
			else
			{
				MessageBox.Show("Firmware binary is missing", "Update Error");
				Close();
			}
		}

		public void UpdateFirmware(ProgressChangedEventHandler update,RunWorkerCompletedEventHandler completed)
        {
			fw.WorkerSupportsCancellation = true;
			fw.WorkerReportsProgress = true;
			fw.DoWork += new DoWorkEventHandler(fw_DoWork);
			fw.ProgressChanged += new ProgressChangedEventHandler(update);
			fw.RunWorkerCompleted += new RunWorkerCompletedEventHandler(completed);
			fw.RunWorkerAsync();
        }

        public bool ST_M1Forward(byte pwr)
        {
            return Write_CRC(new byte[] { m_address, 0, pwr, 0 });
        }

        public bool ST_M2Forward(byte pwr)
        {
			return Write_CRC(new byte[] { m_address, 4, pwr, 0 });
        }

        public bool ST_M1Backward(byte pwr)
        {
			return Write_CRC(new byte[] { m_address, 1, pwr, 0 });
        }

        public bool ST_M2Backward(byte pwr)
        {
			return Write_CRC(new byte[] { m_address, 5, pwr, 0 });
        }

        public bool ST_M1Drive(byte pwr)
        {
			return Write_CRC(new byte[] { m_address, 6, pwr, 0 });
        }

        public bool ST_M2Drive(byte pwr)
        {
			return Write_CRC(new byte[] { m_address, 7, pwr, 0 });
        }

        public bool ST_MixedForward(byte pwr)
        {
			return Write_CRC(new byte[] { m_address, 8, pwr, 0 });
        }

        public bool ST_MixedBackward(byte pwr)
        {
			return Write_CRC(new byte[] { m_address, 9, pwr, 0 });
        }

        public bool ST_MixedLeft(byte pwr)
        {
			return Write_CRC(new byte[] { m_address, 11, pwr, 0 });
        }

        public bool ST_MixedRight(byte pwr)
        {
			return Write_CRC(new byte[] { m_address, 10, pwr, 0 });
        }

        public bool ST_MixedDrive(byte pwr)
        {
			return Write_CRC(new byte[] { m_address, 12, pwr, 0 });
        }

        public bool ST_MixedTurn(byte pwr)
        {
			return Write_CRC(new byte[] { m_address, 13, pwr, 0 });
        }

        public bool ST_SetMinMainVoltage(byte set)
        {
			return Write_CRC(new byte[] { m_address, 2, set, 0 });
        }

        public bool ST_SetMaxMainVoltage(byte set)
        {
			return Write_CRC(new byte[] { m_address, 3, set, 0 });
        }

        public bool GetM1Encoder(out Int32 enc,out byte status)
        {
            byte[] arr;
            if(readcmd(m_address, 16, 5, out arr)){
                enc = (Int32)(arr[0] << 24 | arr[1] << 16 | arr[2] << 8 | arr[3]);
                status = arr[4];
                return true;
            }
            enc = 0;
            status = 0;
            return false;
        }

        public bool GetM2Encoder(out Int32 enc, out byte status)
        {
            byte[] arr;
            if (readcmd(m_address, 17, 5, out arr))
            {
                enc = (Int32)(arr[0] << 24 | arr[1] << 16 | arr[2] << 8 | arr[3]);
                status = arr[4];
                return true;
            }
            enc = 0;
            status = 0;
            return false;
        }

        public bool GetM1Speed(out Int32 speed,out byte status)
        {
            byte[] arr;
            if (readcmd(m_address, 18, 5, out arr))
            {
                speed = (Int32)(arr[0] << 24 | arr[1] << 16 | arr[2] << 8 | arr[3]);
                status = arr[4];
                return true;
            }
            speed = 0;
            status = 0;
            return false;
        }

        public bool GetM2Speed(out Int32 speed, out byte status)
        {
            byte[] arr;
            if (readcmd(m_address, 19, 5, out arr))
            {
                speed = (Int32)(arr[0] << 24 | arr[1] << 16 | arr[2] << 8 | arr[3]);
                status = arr[4];
                return true;
            }
            speed = 0;
            status = 0;
            return false;
        }

        public bool ResetEncoders()
        {
			return Write_CRC(new byte[] { m_address, 20, 0 });
        }

        public bool GetVersion(out string version)
        {
            try
            {
                serialPort.Write(new byte[] { m_address, 21 }, 0, 2);
                version = serialPort.ReadLine();
                serialPort.ReadByte();
                serialPort.ReadByte();
                return true;
            }
            catch (Exception)
            {
                version = "";
            }
            return false;
        }

        public bool GetMainVoltage(out double voltage)
        {
            voltage = 0;
            byte[] arr;
            if (readcmd(m_address, 24, 2, out arr))
            {
                voltage = (double)(arr[0] << 8 | arr[1])/10.0;
                return true;
            }
            return false;
        }

        public bool GetLogicVoltage(out double voltage)
        {
            voltage = 0;
            byte[] arr;
            if (readcmd(m_address, 25, 2, out arr))
            {
                voltage = (double)(arr[0] << 8 | arr[1])/10.0;
                return true;
            }
            return false;
        }

        public bool ST_SetMaxLogicVoltage(byte set)
        {
			return Write_CRC(new byte[] { m_address, 26, set, 0 });
        }

        public bool ST_SetMinLogicVolage(byte set)
        {
			return Write_CRC(new byte[] { m_address, 27, set, 0 });
        }

        public bool SetM1VelocityConstants(double P, double I, double D, UInt32 qpps)
        {
            UInt32 p, i, d;
            p = (UInt32)(P * 65536.0);
            i = (UInt32)(I * 65536.0);
            d = (UInt32)(D * 65536.0);

			return Write_CRC(new byte[] { m_address, 28, (byte)(d >> 24), (byte)(d >> 16), (byte)(d >> 8), (byte)d,
                                          (byte)(p >> 24), (byte)(p >> 16), (byte)(p >> 8), (byte)p,
                                          (byte)(i >> 24), (byte)(i >> 16), (byte)(i >> 8), (byte)i,
                                          (byte)(qpps >> 24), (byte)(qpps >> 16), (byte)(qpps >> 8), (byte)qpps,
                                          0 });
        }

        public bool SetM2VelocityConstants(double P, double I, double D, UInt32 qpps)
        {
            UInt32 p, i, d;
            p = (UInt32)(P * 65536.0);
            i = (UInt32)(I * 65536.0);
            d = (UInt32)(D * 65536.0);

			return Write_CRC(new byte[] { m_address, 29, (byte)(d >> 24), (byte)(d >> 16), (byte)(d >> 8), (byte)d,
                                          (byte)(p >> 24), (byte)(p >> 16), (byte)(p >> 8), (byte)p,
                                          (byte)(i >> 24), (byte)(i >> 16), (byte)(i >> 8), (byte)i,
                                          (byte)(qpps >> 24), (byte)(qpps >> 16), (byte)(qpps >> 8), (byte)qpps,
                                          0 });
        }

        public bool GetM1ISpeed(out Int32 speed, out byte status)
        {
            byte[] arr;
            if (readcmd(m_address, 30, 5, out arr))
            {
                speed = (Int32)(arr[0] << 24 | arr[1] << 16 | arr[2] << 8 | arr[3]);
                status = arr[4];
                return true;
            }
            speed = 0;
            status = 0;
            return false;
        }

        public bool GetM2ISpeed(out Int32 speed, out byte status)
        {
            byte[] arr;
            if (readcmd(m_address, 31, 5, out arr))
            {
                speed = (Int32)(arr[0] << 24 | arr[1] << 16 | arr[2] << 8 | arr[3]);
                status = arr[4];
                return true;
            }
            speed = 0;
            status = 0;
            return false;
        }

        public bool M1Duty(Int32 duty)
        {
			return Write_CRC(new byte[] { m_address, 32, (byte)(duty >> 8), (byte)duty, 0 });
        }

        public bool M2Duty(Int32 duty)
        {
			return Write_CRC(new byte[] { m_address, 33, (byte)(duty >> 8), (byte)duty, 0 });
        }

        public bool MixedDuty(Int16 duty1, Int16 duty2)
        {
			return Write_CRC(new byte[] { m_address, 34, (byte)(duty1 >> 8), (byte)duty1, (byte)(duty2 >> 8), (byte)duty2, 0 });
        }

        public bool M1Speed(Int32 speed)
        {
			return Write_CRC(new byte[] { m_address, 35, (byte)(speed >> 24), (byte)(speed >> 16), (byte)(speed >> 8), (byte)speed, 0 });
        }

        public bool M2Speed(Int32 speed)
        {
			return Write_CRC(new byte[] { m_address, 36, (byte)(speed >> 24), (byte)(speed >> 16), (byte)(speed >> 8), (byte)speed, 0 });
        }

        public bool MixedSpeed(Int32 speed1, Int32 speed2)
        {
			return Write_CRC(new byte[] { m_address, 37, (byte)(speed1 >> 24), (byte)(speed1 >> 16), (byte)(speed1 >> 8), (byte)speed1,
                                             (byte)(speed2 >> 24), (byte)(speed2 >> 16), (byte)(speed2 >> 8), (byte)speed2,
                                             0 });
        }

        public bool M1SpeedAccel(UInt32 accel, Int32 speed)
        {
			return Write_CRC(new byte[] { m_address, 38, (byte)(accel >> 24), (byte)(accel >> 16), (byte)(accel >> 8), (byte)accel,
                                             (byte)(speed >> 24), (byte)(speed >> 16), (byte)(speed >> 8), (byte)speed,
                                             0 });
        }

        public bool M2SpeedAccel(UInt32 accel, Int32 speed)
        {
			return Write_CRC(new byte[] { m_address, 39, (byte)(accel >> 24), (byte)(accel >> 16), (byte)(accel >> 8), (byte)accel,
                                             (byte)(speed >> 24), (byte)(speed >> 16), (byte)(speed >> 8), (byte)speed,
                                             0 });
        }

        public bool MixedSpeedAccel(UInt32 accel, Int32 speed1, Int32 speed2)
        {
			return Write_CRC(new byte[] { m_address, 40, (byte)(accel >> 24), (byte)(accel >> 16), (byte)(accel >> 8), (byte)accel,
                                             (byte)(speed1 >> 24), (byte)(speed1 >> 16), (byte)(speed1 >> 8), (byte)speed1,
                                             (byte)(speed2 >> 24), (byte)(speed2 >> 16), (byte)(speed2 >> 8), (byte)speed2,
                                             0 });
        }

        public bool M1SpeedDistance(Int32 speed, UInt32 distance, byte buffer)
        {
			return Write_CRC(new byte[] { m_address, 41, (byte)(speed >> 24), (byte)(speed >> 16), (byte)(speed >> 8), (byte)speed,
                                             (byte)(distance >> 24), (byte)(distance >> 16), (byte)(distance >> 8), (byte)distance,
                                             buffer, 0 });
        }

        public bool M2SpeedDistance(Int32 speed, UInt32 distance, byte buffer)
        {
			return Write_CRC(new byte[] { m_address, 42, (byte)(speed >> 24), (byte)(speed >> 16), (byte)(speed >> 8), (byte)speed,
                                             (byte)(distance >> 24), (byte)(distance >> 16), (byte)(distance >> 8), (byte)distance,
                                             buffer, 0 });
        }

        public bool MixedSpeedDistance(Int32 speed1, UInt32 distance1, Int32 speed2, UInt32 distance2, byte buffer)
        {
			return Write_CRC(new byte[] { m_address, 43, (byte)(speed1 >> 24), (byte)(speed1 >> 16), (byte)(speed1 >> 8), (byte)speed1,
                                             (byte)(distance1 >> 24), (byte)(distance1 >> 16), (byte)(distance1 >> 8), (byte)distance1,
                                             (byte)(speed2 >> 24), (byte)(speed2 >> 16), (byte)(speed2 >> 8), (byte)speed2,
                                             (byte)(distance2 >> 24), (byte)(distance1 >> 16), (byte)(distance2 >> 8), (byte)distance2,
                                             buffer, 0 });
        }

        public bool M1SpeedAccelDistance(Int32 accel, UInt32 speed, UInt32 distance, byte buffer)
        {
			return Write_CRC(new byte[] { m_address, 44, (byte)(accel >> 24), (byte)(accel >> 16), (byte)(accel >> 8), (byte)accel,
                                             (byte)(speed >> 24), (byte)(speed >> 16), (byte)(speed >> 8), (byte)speed,
                                             (byte)(distance >> 24), (byte)(distance >> 16), (byte)(distance >> 8), (byte)distance,
                                             buffer, 0 });
        }

        public bool M2SpeedAccelDistance(Int32 accel, UInt32 speed, UInt32 distance, byte buffer)
        {
			return Write_CRC(new byte[] { m_address, 45, (byte)(accel >> 24), (byte)(accel >> 16), (byte)(accel >> 8), (byte)accel,
                                             (byte)(speed >> 24), (byte)(speed >> 16), (byte)(speed >> 8), (byte)speed,
                                             (byte)(distance >> 24), (byte)(distance >> 16), (byte)(distance >> 8), (byte)distance,
                                             buffer, 0 });
        }

        public bool MixedSpeedAccelDistance(UInt32 accel, Int32 speed1, UInt32 distance1, Int32 speed2, UInt32 distance2, byte buffer)
        {
			return Write_CRC(new byte[] { m_address, 46, (byte)(accel >> 24), (byte)(accel >> 16), (byte)(accel >> 8), (byte)accel,
                                             (byte)(speed1 >> 24), (byte)(speed1 >> 16), (byte)(speed1 >> 8), (byte)speed1,
                                             (byte)(distance1 >> 24), (byte)(distance1 >> 16), (byte)(distance1 >> 8), (byte)distance1,
                                             (byte)(speed2 >> 24), (byte)(speed2 >> 16), (byte)(speed2 >> 8), (byte)speed2,
                                             (byte)(distance2 >> 24), (byte)(distance1 >> 16), (byte)(distance2 >> 8), (byte)distance2,
                                             buffer, 0 });
        }

        public bool GetBuffers(out byte buffer1, out byte buffer2)
        {
            byte[] arr;
            if (readcmd(m_address, 47, 2, out arr))
            {
                buffer1 = arr[0];
                buffer2 = arr[1];
                return true;
            }
            buffer1 = buffer2 = 0;
            return false;
        }

        public bool GetCurrents(out Int16 current1, out Int16 current2)
        {
            byte[] arr;
            if (readcmd(m_address, 49, 4, out arr))
            {
                current1 = (Int16)(arr[0] << 8 | arr[1]);
                current2 = (Int16)(arr[2] << 8 | arr[3]);
                return true;
            }
            current1 = current2 = 0;
            return false;
        }

        public bool MixedSpeedAccel2(Int32 accel1, UInt32 speed1, Int32 accel2, UInt32 speed2)
        {
			return Write_CRC(new byte[] { m_address, 50, (byte)(accel1 >> 24), (byte)(accel1 >> 16), (byte)(accel1 >> 8), (byte)accel1,
                                             (byte)(speed1 >> 24), (byte)(speed1 >> 16), (byte)(speed1 >> 8), (byte)speed1,
                                             (byte)(accel2 >> 24), (byte)(accel2 >> 16), (byte)(accel2 >> 8), (byte)accel2,
                                             (byte)(speed2 >> 24), (byte)(speed2 >> 16), (byte)(speed2 >> 8), (byte)speed2,
                                             0 });
        }

        public bool MixedSpeedAccelDistance2(Int32 accel1, UInt32 speed1, UInt32 distance1, Int32 accel2, UInt32 speed2, UInt32 distance2, byte buffer)
        {
			return Write_CRC(new byte[] { m_address, 51, (byte)(accel1 >> 24), (byte)(accel1 >> 16), (byte)(accel1 >> 8), (byte)accel1,
                                             (byte)(speed1 >> 24), (byte)(speed1 >> 16), (byte)(speed1 >> 8), (byte)speed1,
                                             (byte)(distance1 >> 24), (byte)(distance1 >> 16), (byte)(distance1 >> 8), (byte)distance1,
                                             (byte)(accel2 >> 24), (byte)(accel2 >> 16), (byte)(accel2 >> 8), (byte)accel2,
                                             (byte)(speed2 >> 24), (byte)(speed2 >> 16), (byte)(speed2 >> 8), (byte)speed2,
                                             (byte)(distance2 >> 24), (byte)(distance1 >> 16), (byte)(distance2 >> 8), (byte)distance2,
                                             buffer, 0 });
        }

        public bool M1DutyAccel(Int16 duty, UInt16 accel)
        {
			return Write_CRC(new byte[] { m_address, 52, (byte)(duty >> 8), (byte)duty,
                                             (byte)(accel >> 8), (byte)accel,
                                             0 });
        }

        public bool M2DutyAccel(Int16 duty, UInt16 accel)
        {
			return Write_CRC(new byte[] { m_address, 53, (byte)(duty >> 8), (byte)duty,
                                             (byte)(accel >> 8), (byte)accel,
                                             0 });
        }

        public bool MixedDutyAccel(Int16 duty1, UInt16 accel1, Int16 duty2, UInt16 accel2)
        {
			return Write_CRC(new byte[] { m_address, 54, (byte)(duty1 >> 8), (byte)duty1,
                                             (byte)(accel1 >> 8), (byte)accel1,
                                             (byte)(duty2 >> 8), (byte)duty2,
                                             (byte)(accel2 >> 8), (byte)accel2,
                                             0 });
        }

        public bool GetM1VelocityConstants(out double p, out double i, out double d, out UInt32 qpps)
        {
            byte[] arr;
            if (readcmd(m_address, 55, 16, out arr))
            {
                p = (double)(arr[0] << 24 | arr[1] << 16 | arr[2] << 8 | arr[3]) / 65536;
                i = (double)(arr[4] << 24 | arr[5] << 16 | arr[6] << 8 | arr[7]) / 65536;
                d = (double)(arr[8] << 24 | arr[9] << 16 | arr[10] << 8 | arr[11]) / 65536;
                qpps = (UInt32)(arr[12] << 24 | arr[13] << 16 | arr[14] << 8 | arr[15]);
                return true;
            }
            p = i = d = qpps = 0;
            return false;
        }

        public bool GetM2VelocityConstants(out double p, out double i, out double d, out UInt32 qpps)
        {
            byte[] arr;
            if (readcmd(m_address, 56, 16, out arr))
            {
                p = (double)(arr[0] << 24 | arr[1] << 16 | arr[2] << 8 | arr[3]) / 65536;
                i = (double)(arr[4] << 24 | arr[5] << 16 | arr[6] << 8 | arr[7]) / 65536;
                d = (double)(arr[8] << 24 | arr[9] << 16 | arr[10] << 8 | arr[11]) / 65536;
                qpps = (UInt32)(arr[12] << 24 | arr[13] << 16 | arr[14] << 8 | arr[15]);
                return true;
            }
            p = i = d = qpps = 0;
            return false;
        }

        public bool SetMainVoltageLimits(double Min, double Max)
        {
            UInt16 min = (UInt16)(Min * 10.0);
            UInt16 max = (UInt16)(Max * 10.0);
			return Write_CRC(new byte[] { m_address, 57, (byte)(min >> 8), (byte)min,
                                             (byte)(max >> 8), (byte)max,
                                             0 });
        }

        public bool SetLogicVoltageLimits(double Min, double Max)
        {
            UInt16 min = (UInt16)(Min * 10.0);
            UInt16 max = (UInt16)(Max * 10.0);
			return Write_CRC(new byte[] { m_address, 58, (byte)(min >> 8), (byte)min,
                                             (byte)(max >> 8), (byte)max,
                                             0 });
        }

        public bool GetMainVoltageLimits(out double min, out double max)
        {
            byte[] arr;
            if (readcmd(m_address, 59, 4, out arr))
            {
                min = (double)(arr[0] << 8 | arr[1])/10.0;
                max = (double)(arr[2] << 8 | arr[3])/10.0;
                return true;
            }
            min = max = 0;
            return false;
        }

        public bool GetLogicVoltageLimits(out double min, out double max)
        {
            byte[] arr;
            if (readcmd(m_address, 60, 4, out arr))
            {
                min = (double)(arr[0] << 8 | arr[1])/10.0;
                max = (double)(arr[2] << 8 | arr[3])/10.0;
                return true;
            }
            min = max = 0;
            return false;
        }

        public bool SetM1PositionConstants(double P, double I, double D, UInt32 imax, UInt32 deadzone, Int32 minlimit, Int32 maxlimit)
        {
            UInt32 p, i, d;
            p = (UInt32)(P * 1024.0);
            i = (UInt32)(I * 1024.0);
            d = (UInt32)(D * 1024.0);

			return Write_CRC(new byte[] { m_address, 61, (byte)(d >> 24), (byte)(d >> 16), (byte)(d >> 8), (byte)d,
                                          (byte)(p >> 24), (byte)(p >> 16), (byte)(p >> 8), (byte)p,
                                          (byte)(i >> 24), (byte)(i >> 16), (byte)(i >> 8), (byte)i,
                                          (byte)(imax >> 24), (byte)(imax >> 16), (byte)(imax >> 8), (byte)imax,
                                          (byte)(deadzone >> 24), (byte)(deadzone >> 16), (byte)(deadzone >> 8), (byte)deadzone,
                                          (byte)(minlimit >> 24), (byte)(minlimit >> 16), (byte)(minlimit >> 8), (byte)minlimit,
                                          (byte)(maxlimit >> 24), (byte)(maxlimit >> 16), (byte)(maxlimit >> 8), (byte)maxlimit,
                                          0 });
        }

        public bool SetM2PositionConstants(double P, double I, double D, UInt32 imax, UInt32 deadzone, Int32 minlimit, Int32 maxlimit)
        {
            UInt32 p, i, d;
            p = (UInt32)(P * 1024.0);
            i = (UInt32)(I * 1024.0);
            d = (UInt32)(D * 1024.0);

			return Write_CRC(new byte[] { m_address, 62, (byte)(d >> 24), (byte)(d >> 16), (byte)(d >> 8), (byte)d,
                                          (byte)(p >> 24), (byte)(p >> 16), (byte)(p >> 8), (byte)p,
                                          (byte)(i >> 24), (byte)(i >> 16), (byte)(i >> 8), (byte)i,
                                          (byte)(imax >> 24), (byte)(imax >> 16), (byte)(imax >> 8), (byte)imax,
                                          (byte)(deadzone >> 24), (byte)(deadzone >> 16), (byte)(deadzone >> 8), (byte)deadzone,
                                          (byte)(minlimit >> 24), (byte)(minlimit >> 16), (byte)(minlimit >> 8), (byte)minlimit,
                                          (byte)(maxlimit >> 24), (byte)(maxlimit >> 16), (byte)(maxlimit >> 8), (byte)maxlimit,
                                          0 });
        }

        public bool GetM1PositionConstants(out double p, out double i, out double d, out UInt32 imax, out UInt32 deadzone, out Int32 minlimit, out Int32 maxlimit)
        {
            byte[] arr;
            if (readcmd(m_address, 63, 28, out arr))
            {
                p = (double)(arr[0] << 24 | arr[1] << 16 | arr[2] << 8 | arr[3]) / 1024;
                i = (double)(arr[4] << 24 | arr[5] << 16 | arr[6] << 8 | arr[7]) / 1024;
                d = (double)(arr[8] << 24 | arr[9] << 16 | arr[10] << 8 | arr[11]) / 1024;
                imax = (UInt32)(arr[12] << 24 | arr[13] << 16 | arr[14] << 8 | arr[15]);
                deadzone = (UInt32)(arr[16] << 24 | arr[17] << 16 | arr[18] << 8 | arr[19]);
                minlimit = (Int32)(arr[20] << 24 | arr[21] << 16 | arr[22] << 8 | arr[23]);
                maxlimit = (Int32)(arr[24] << 24 | arr[25] << 16 | arr[26] << 8 | arr[27]);
                return true;
            }
            p = i = d = imax = deadzone = 0;
            minlimit = maxlimit = 0;
            return false;
        }

        public bool GetM2PositionConstants(out double p, out double i, out double d, out UInt32 imax, out UInt32 deadzone, out Int32 minlimit, out Int32 maxlimit)
        {
            byte[] arr;
            if (readcmd(m_address, 64, 28, out arr))
            {
                p= (double)(arr[0] << 24 | arr[1] << 16 | arr[2] << 8 | arr[3]) / 1024;
                i = (double)(arr[4] << 24 | arr[5] << 16 | arr[6] << 8 | arr[7]) / 1024;
                d = (double)(arr[8] << 24 | arr[9] << 16 | arr[10] << 8 | arr[11]) / 1024;
                imax = (UInt32)(arr[12] << 24 | arr[13] << 16 | arr[14] << 8 | arr[15]);
                deadzone = (UInt32)(arr[16] << 24 | arr[17] << 16 | arr[18] << 8 | arr[19]);
                minlimit = (Int32)(arr[20] << 24 | arr[21] << 16 | arr[22] << 8 | arr[23]);
                maxlimit = (Int32)(arr[24] << 24 | arr[25] << 16 | arr[26] << 8 | arr[27]);
                return true;
            }
            p = i = d = imax = deadzone = 0;
            minlimit = maxlimit = 0;
            return false;
        }

        public bool M1SpeedAccelDeccelPosition(UInt32 accel, UInt32 speed, UInt32 deccel, Int32 position, byte buffer)
        {
            return Write_CRC(new byte[] { m_address, 65, (byte)(accel >> 24), (byte)(accel >> 16), (byte)(accel >> 8), (byte)accel,
                                             (byte)(speed >> 24), (byte)(speed >> 16), (byte)(speed >> 8), (byte)speed,
                                             (byte)(deccel >> 24), (byte)(deccel >> 16), (byte)(deccel >> 8), (byte)deccel,
                                             (byte)(position >> 24), (byte)(position >> 16), (byte)(position >> 8), (byte)position,
                                             buffer, 0 });
        }

        public bool M2SpeedAccelDeccelPosition(UInt32 accel, UInt32 speed, UInt32 deccel, Int32 position, byte buffer)
        {
            return Write_CRC(new byte[] { m_address, 66, (byte)(accel >> 24), (byte)(accel >> 16), (byte)(accel >> 8), (byte)accel,
                                             (byte)(speed >> 24), (byte)(speed >> 16), (byte)(speed >> 8), (byte)speed,
                                             (byte)(deccel >> 24), (byte)(deccel >> 16), (byte)(deccel >> 8), (byte)deccel,
                                             (byte)(position >> 24), (byte)(position >> 16), (byte)(position >> 8), (byte)position,
                                             buffer, 0 });
        }

        public bool MixedSpeedAccelDeccelPosition(UInt32 accel1, UInt32 speed1, UInt32 deccel1, Int32 position1, UInt32 accel2, UInt32 speed2, UInt32 deccel2, Int32 position2, byte buffer)
        {
            return Write_CRC(new byte[] { m_address, 67, (byte)(accel1 >> 24), (byte)(accel1 >> 16), (byte)(accel1 >> 8), (byte)accel1,
                                             (byte)(speed1 >> 24), (byte)(speed1 >> 16), (byte)(speed1 >> 8), (byte)speed1,
                                             (byte)(deccel1 >> 24), (byte)(deccel1 >> 16), (byte)(deccel1 >> 8), (byte)deccel1,
                                             (byte)(position1 >> 24), (byte)(position1 >> 16), (byte)(position1 >> 8), (byte)position1,
                                             (byte)(accel2 >> 24), (byte)(accel2 >> 16), (byte)(accel2 >> 8), (byte)accel2,
                                             (byte)(speed2 >> 24), (byte)(speed2 >> 16), (byte)(speed2 >> 8), (byte)speed2,
                                             (byte)(deccel2 >> 24), (byte)(deccel2 >> 16), (byte)(deccel2 >> 8), (byte)deccel2,
                                             (byte)(position2 >> 24), (byte)(position2 >> 16), (byte)(position2 >> 8), (byte)position2,
                                             buffer, 0 });
        }

        public bool GetTemperature(out double temperature)
        {
            byte[] arr;
            if (readcmd(m_address, 82, 2, out arr))
            {
                temperature = (double)(arr[0] << 8 | arr[1])/10.0;
                return true;
            }
            temperature = 0;
            return false;
        }

		public bool GetTemperature2(out double temperature2)
		{
			byte[] arr;
			if (readcmd(m_address, 83, 2, out arr))
			{
				temperature2 = (double)(arr[0] << 8 | arr[1]) / 10.0;
				return true;
			}
			temperature2 = 0;
			return false;
		}

		public bool SetDitherM1(byte rate, Int16 offset)
		{
			return Write_CRC(new byte[] { m_address, 70, (byte)rate,
                                             (byte)(offset >> 8), (byte)offset,
                                             0 });
		}

		public bool SetDitherM2(byte rate, Int16 offset)
		{
			return Write_CRC(new byte[] { m_address, 71, (byte)rate,
                                             (byte)(offset >> 8), (byte)offset,
                                             0 });
		}

		public bool GetDitherM1(out byte rate, out Int16 offset)
		{
			byte[] arr;
			if (readcmd(m_address, 72, 3, out arr))
			{
				rate = arr[0];
				offset = (Int16)((arr[1] << 8) | arr[2]);
				return true;
			}
			rate = 0;
			offset = 0;
			return false;
		}

		public bool GetDitherM2(out byte rate, out Int16 offset)
		{
			byte[] arr;
			if (readcmd(m_address, 73, 3, out arr))
			{
				rate = arr[0];
				offset = (Int16)((arr[1] << 8) | arr[2]);
				return true;
			}
			rate = 0;
			offset = 0;
			return false;
		}

		public bool VelocityAutotuneM1(UInt32 amplitude, Int32 targetspeed)
        {
			return Write_CRC(new byte[] { m_address, 84, (byte)(amplitude >> 24), (byte)(amplitude >> 16), (byte)(amplitude >> 8), (byte)amplitude,
                                             (byte)(targetspeed >> 24), (byte)(targetspeed >> 16), (byte)(targetspeed >> 8), (byte)targetspeed,
                                             0 });
        }

		public bool VelocityAutotuneM2(UInt32 amplitude, Int32 targetspeed)
        {
			return Write_CRC(new byte[] { m_address, 85, (byte)(amplitude >> 24), (byte)(amplitude >> 16), (byte)(amplitude >> 8), (byte)amplitude,
                                             (byte)(targetspeed >> 24), (byte)(targetspeed >> 16), (byte)(targetspeed >> 8), (byte)targetspeed,
                                             0 });
        }

		public bool PositionAutotuneM1(UInt32 amplitude, Int32 targetpos)
        {
			return Write_CRC(new byte[] { m_address, 86, (byte)(amplitude >> 24), (byte)(amplitude >> 16), (byte)(amplitude >> 8), (byte)amplitude,
                                             (byte)(targetpos >> 24), (byte)(targetpos >> 16), (byte)(targetpos >> 8), (byte)targetpos,
                                             0 });
        }

		public bool PositionAutotuneM2(UInt32 amplitude, Int32 targetpos)
        {
			return Write_CRC(new byte[] { m_address, 87, (byte)(amplitude >> 24), (byte)(amplitude >> 16), (byte)(amplitude >> 8), (byte)amplitude,
                                             (byte)(targetpos >> 24), (byte)(targetpos >> 16), (byte)(targetpos >> 8), (byte)targetpos,
                                             0 });
        }

        public void GetM1Data(out Int32[] data)
        {
            data = new Int32[32];
            byte[] arr;
            serialPort.ReadTimeout = 20000;
            if (readcmd(m_address, 88, 128, out arr))
            {
                for (int i = 0; i < 32; i++)
                {
                    data[i] = (Int32)(arr[(i * 4) + 0] << 24 | arr[(i * 4) + 1] << 16 | arr[(i * 4) + 2] << 8 | arr[(i * 4) + 3]);
                }
            }
            serialPort.ReadTimeout = 100;
        }

        public void GetM2Data(out Int32[] data)
        {
            data = new Int32[32];
            byte[] arr;
            serialPort.ReadTimeout = 20000;
            if (readcmd(m_address, 89, 128, out arr))
            {
                for (int i = 0; i < 32; i++)
                {
                    data[i] = (Int32)(arr[(i * 4) + 0] << 24 | arr[(i * 4) + 1] << 16 | arr[(i * 4) + 2] << 8 | arr[(i * 4) + 3]);
                }
            }
            serialPort.ReadTimeout = 100;
        }

        public bool GetStatus(out UInt16 status)
        {
            byte[] arr;
            if (readcmd(m_address, 90, 2, out arr))
            {
                status = (UInt16)(arr[0]<<8 | arr[1]);
                return true;
            }
            status = 0;
            return false;
        }

        public bool GetEncoderModes(out byte m1mode, out byte m2mode)
        {
            byte[] arr;
            if (readcmd(m_address, 91, 2, out arr))
            {
                m1mode = arr[0];
                m2mode = arr[1];
                return true;
            }
            m1mode = m2mode = 0;
            return false;
        }

        public bool SetEncoder1Mode(byte mode)
        {
			return Write_CRC(new byte[] { m_address, 92, mode, 0 });
        }

        public bool SetEncoder2Mode(byte mode)
        {
			return Write_CRC(new byte[] { m_address, 93, mode, 0 });
        }

		public bool WriteNVM()
        {
			return Write_CRC(new byte[] { m_address, 94, 0 });
        }

		public bool ReadNVM()
		{
			return Write_CRC(new byte[] { m_address, 95, 0 });
		}

		public bool SetConfig(UInt16 config)
        {
			byte oldaddress = m_address;
			m_address = (byte)(((config >> 8) & 0x07) + 0x80);
			return Write_CRC(new byte[] { oldaddress, 98, (byte)(config >> 8), (byte)config, 0 });
        }

        public bool GetConfig(out UInt16 config)
        {
            byte[] arr;
            if (readcmd(m_address, 99, 2, out arr))
            {
                config = (UInt16)(arr[0] << 8 | arr[1]);
                return true;
            }
            config = 0;
            return false;
        }

		public void GetData(out Int16[] data)
		{
			data = new Int16[256];
			byte[] arr;
			serialPort.ReadTimeout = 20000;
			if (readcmd(m_address, 96, 512, out arr))
			{
				for (UInt16 i = 0; i < 256; i++)
				{
					data[i] = (Int16)(arr[(i * 2) + 0] << 8 | arr[(i * 2) + 1]);
				}
			}
			serialPort.ReadTimeout = 100;
		}

	}
}
